#include <Arduino.h>
#include <FS.h>                   //this needs to be first, or it all crashes and burns...
#include <ESP8266WiFi.h>          //https://github.com/esp8266/Arduino

//needed for library
#include <DNSServer.h>
#include <ESP8266WebServer.h>
#include <WiFiManager.h>

#include <ArduinoJson.h>
#include <SPI.h>
#include <RF24.h>

#include "config.h"
#include "display.h"
#include "fastclock.h"

#define AUTOSTART_CLOCK_AFTER_ms 10000

//define your default values here, if there are different values in config.json, they are overwritten.
//length should be max size + 1
char mqtt_server[40] = "";
char mqtt_port[6] = "8080";
//char blynk_token[33] = "YOUR_BLYNK_TOKEN";
//default custom static IP
char static_ip[16] = "10.0.1.56";
char static_gw[16] = "10.0.1.1";
char static_sn[16] = "255.255.255.0";

//flag for saving data
bool shouldSaveConfig = false;

RF24 radio(PIN_NRF24_CE, PIN_NRF24_CSN);
byte addresses[][6] = {NETWORK_ADDRESS_MASTER_SEND, NETWORK_ADDRESS_MASTER_RECEIVE};

// FastClock
bool isMaster = true;

Display display;
Fastclock fastclock(&display);

//callback notifying us of the need to save config
void saveConfigCallback () {
  Serial.println("Should save config");
  shouldSaveConfig = true;
}

void setupWifiConnection() {
  // The extra parameters to be configured (can be either global or just in the setup)
  // After connecting, parameter.getValue() will get you the configured value
  // id/name placeholder/prompt default length
  //**WiFiManagerParameter custom_mqtt_server("server", "mqtt server", mqtt_server, 40);
  //**WiFiManagerParameter custom_mqtt_port("port", "mqtt port", mqtt_port, 5);
  //WiFiManagerParameter custom_blynk_token("blynk", "blynk token", blynk_token, 34);

  //WiFiManager
  //Local intialization. Once its business is done, there is no need to keep it around
  WiFiManager wifiManager;

  //set config save notify callback
  wifiManager.setSaveConfigCallback(saveConfigCallback);

  //set static ip
  IPAddress _ip,_gw,_sn;
  _ip.fromString(static_ip);
  _gw.fromString(static_gw);
  _sn.fromString(static_sn);

  //wifiManager.setSTAStaticIPConfig(_ip, _gw, _sn);

  //add all your parameters here
  //**wifiManager.addParameter(&custom_mqtt_server);
  //**wifiManager.addParameter(&custom_mqtt_port);
  //wifiManager.addParameter(&custom_blynk_token);

  //reset settings - for testing
  //wifiManager.resetSettings();

  //set minimu quality of signal so it ignores AP's under that quality
  //defaults to 8%
  wifiManager.setMinimumSignalQuality();

  //sets timeout until configuration portal gets turned off
  //useful to make it all retry or go to sleep
  //in seconds
  //wifiManager.setTimeout(120);

  //fetches ssid and pass and tries to connect
  //if it does not connect it starts an access point with the specified name
  //and goes into a blocking loop awaiting configuration
  Serial.println("Starting autoConnect ...");
  if (!wifiManager.autoConnect("FastclockMasterAP", "password")) {
    Serial.println("failed to connect and hit timeout");
    delay(3000);
    //reset and try again, or maybe put it to deep sleep
    ESP.reset();
    delay(5000);
  }

  //if you get here you have connected to the WiFi
  Serial.println("connected...yeey :)");
  display.addLogMessage("WLAN connected");

  //read updated parameters
  //**strcpy(mqtt_server, custom_mqtt_server.getValue());
  //**strcpy(mqtt_port, custom_mqtt_port.getValue());
  //strcpy(blynk_token, custom_blynk_token.getValue());

  //save the custom parameters to FS
  if (shouldSaveConfig) {
    Serial.println("saving config");
    DynamicJsonBuffer jsonBuffer;
    JsonObject& json = jsonBuffer.createObject();
    //**json["mqtt_server"] = mqtt_server;
    //**json["mqtt_port"] = mqtt_port;
    //json["blynk_token"] = blynk_token;

    json["ip"] = WiFi.localIP().toString();
    json["gateway"] = WiFi.gatewayIP().toString();
    json["subnet"] = WiFi.subnetMask().toString();

    File configFile = SPIFFS.open("/config.json", "w");
    if (!configFile) {
      Serial.println("failed to open config file for writing");
    }

    json.prettyPrintTo(Serial);
    json.printTo(configFile);
    configFile.close();
    //end save
  }

  Serial.print("local ip: "); Serial.println(WiFi.localIP());
  Serial.print("gateway: "); Serial.println(WiFi.gatewayIP());
  Serial.print("subnet: "); Serial.println(WiFi.subnetMask());
}

void setup() {
  // put your setup code here, to run once:
  Serial.begin(115200);
  Serial.println("Starting *** FastclockMasterESP8266");
  display.begin();
  display.showLog();

  //clean FS, for testing
  //SPIFFS.format();

  //read configuration from FS json
  Serial.println("mounting FS...");
  display.addLogMessage("mounting FS ...");

  if (SPIFFS.begin()) {
    Serial.println("mounted file system");
    display.addLogMessage("mounting FS done");
    if (SPIFFS.exists("/config.json")) {
      //file exists, reading and loading
      Serial.println("reading config file");
      display.addLogMessage("reading config file");
      File configFile = SPIFFS.open("/config.json", "r");
      if (configFile) {
        Serial.println("opened config file");
        size_t size = configFile.size();
        // Allocate a buffer to store contents of the file.
        std::unique_ptr<char[]> buf(new char[size]);

        configFile.readBytes(buf.get(), size);
        DynamicJsonBuffer jsonBuffer;
        JsonObject& json = jsonBuffer.parseObject(buf.get());
        json.printTo(Serial);
        if (json.success()) {
          Serial.println("\nparsed json");

          //**strcpy(mqtt_server, json["mqtt_server"]);
          //**strcpy(mqtt_port, json["mqtt_port"]);
          //strcpy(blynk_token, json["blynk_token"]);

          if (json["ip"]) {
            Serial.print("setting custom ip from config: ");
            //**strcpy(static_ip, json["ip"]);
            //**strcpy(static_gw, json["gateway"]);
            //**strcpy(static_sn, json["subnet"]);
            Serial.println(static_ip);
/*            Serial.println("converting ip");
            IPAddress ip = ipFromCharArray(static_ip);
            Serial.println(ip);*/
          } else {
            Serial.println("no custom ip in config");
          }
        } else {
          display.addLogMessage("failed loading json");
          Serial.println("failed to load json config");
        }
      }
    }
  } else {
    display.addLogMessage("failed mounting FS");
    Serial.println("failed to mount FS");
  }
  //end read
  display.addLogMessage(static_ip);
  display.addLogMessage(mqtt_server);
  Serial.print("static ip: "); Serial.println(static_ip);
  //Serial.println(blynk_token);
  //**Serial.print("mqtt: "); Serial.println(mqtt_server);

  // setupWifiConnection();

  // setting up RF24
  #if 0
  display.addLogMessage("Start RF24 radio");
  radio.begin();
  // Open a writing and reading pipe on each radio, with opposite addresses
  if (isMaster){
    radio.openWritingPipe(addresses[1]);
    radio.openReadingPipe(1,addresses[0]);
  } else {
    radio.openWritingPipe(addresses[0]);
    radio.openReadingPipe(1,addresses[1]);
  }

  // Start the radio listening for data
  radio.startListening();
  #endif

  fastclock.begin();
}

void loop(void)
{
  if (!display.showBootSequenceFinished(1000)) {
    return;
  }

#if defined(AUTOSTART_CLOCK_AFTER_ms)
#if AUTOSTART_CLOCK_AFTER_ms > 0
  static unsigned long autostart_ms = 0;
  static bool autostartDone = false;
  if (autostart_ms == 0) {
    autostart_ms = millis() + AUTOSTART_CLOCK_AFTER_ms;
  } else if (!autostartDone && millis() > autostart_ms) {
    autostartDone = true;
    fastclock.setClockHalted(false);
    Serial.println("Clock started automatically");
  }
#endif
#endif

  fastclock.loop();
  display.showDashboard();

  #if 0
  // put your main code here, to run repeatedly:
  /****************** Ping Out Role ***************************/
  if (isMaster)  {

      radio.stopListening();

      Serial.println(F("Now sending"));

      unsigned long start_time = micros();
       if (!radio.write(&start_time, sizeof(unsigned long))) {
         Serial.println(F("failed"));
       }

      radio.startListening();

      unsigned long started_waiting_at = micros();
      boolean timeout = false;

      while (!radio.available()) {
        if (micros() - started_waiting_at > 200000 ) {            // If waited longer than 200ms, indicate timeout and exit while loop
            timeout = true;
            break;
        }
      }

      if (timeout) {
          Serial.println(F("Failed, response timed out."));
      } else {
          unsigned long got_time;
          radio.read( &got_time, sizeof(unsigned long) );
          unsigned long end_time = micros();

          // Spew it
          Serial.print(F("Sent "));
          Serial.print(start_time);
          Serial.print(F(", Got response "));
          Serial.print(got_time);
          Serial.print(F(", Round-trip delay "));
          Serial.print(end_time-start_time);
          Serial.println(F(" microseconds"));
      }

      // Try again 1s later
      delay(1000);
    }



  /****************** Pong Back Role ***************************/

    if (!isMaster)
    {
      unsigned long got_time;

      if( radio.available()){
                                                                      // Variable for the received timestamp
        while (radio.available()) {                                   // While there is data ready
          radio.read( &got_time, sizeof(unsigned long) );             // Get the payload
        }

        radio.stopListening();                                        // First, stop listening so we can talk
        radio.write( &got_time, sizeof(unsigned long) );              // Send the final one back.
        radio.startListening();                                       // Now, resume listening so we catch the next packets.
        Serial.print(F("Sent response "));
        Serial.println(got_time);
     }
   }
   #endif
}


#if 0
#ifdef RF_RadioHead
  if (!Datagram.init()) {
    LOG(LL_ERROR, ("*** Datagram init failed"));
  }
#endif

#ifdef RF_RF24
  radio.begin();
  if (radio.isChipConnected()) { LOG(LL_INFO, ("*** RF chip found")); }
  else { LOG(LL_ERROR, ("*** ERROR: RF chip not found!")); }
  radio.setChannel(1);
  radio.setPALevel(RF24_PA_MAX);
  radio.setDataRate(RF24_2MBPS);
  radio.setAutoAck(0);
  //radio.setRetries(2,15); // Optionally, increase the delay between retries & # of retries
  radio.setCRCLength(RF24_CRC_8);
  radio.openWritingPipe(pipes[0]);
  radio.openReadingPipe(1,pipes[1]);
  radio.startListening();
  radio.printDetails();
  // @TODO: real random seed!
  //randomSeed(analogRead(0));
  //randomSeed(22);
  radio.powerUp();
  LOG(LL_INFO, ("*** RF payload size=%d bytes", radio.getPayloadSize()));
  if (radio.testCarrier() || radio.testRPD()) { LOG(LL_INFO, ("*** Carrier/RPD seen on radio")); }
  if (radio.failureDetected) { LOG(LL_ERROR, ("*** Radio error detected!")); }
#endif
#endif
